/**************************************************************************************

        PROTUX - THE FREE PROFESSIONAL AUDIO TOOLS FOR LINUX
        AUTHOR : See AUTHORS file for details

        This software is distributed under the terms of the GNU General Public License
        as specified in the COPYING file.

***************************************************************************************/
#include <sys/stat.h>
#include <sys/types.h>
#include <math.h>
#include <errno.h>

#include "Song.hh"
#include "BusMonitor.hh"


#include <qsplitter.h>
#include <qfiledialog.h>
#include <qtextstream.h>
#include <qvaluelist.h>
#include <mustux.h>
#include <qdatetime.h>

static const long long PLAY_BLOCKS_OFFSET = 0;


Song::Song(int pMyNumber, Project* pParentProject, QWidget* parent) : QWidget(parent)
        {
        PENTERCONS;
        myNumber = pMyNumber;
        QString sMyNumber = ""; sMyNumber.setNum( myNumber );
        parentProject=pParentProject;
        audioSourcesList=new AudioSourcesList();

        editingMode = EDITING_MODE_NORMAL;
        numTracks = GlobalProperties::get_int("DEFAULT_NUM_TRACKS");
        if (numTracks > MAX_TRACKS)
                {
                PERROR("Internal error. Trying to create %d . Using default MAX TRACKS = %d", numTracks, MAX_TRACKS );
                numTracks = MAX_TRACKS;
                }
        activeTrackNumber=0;
        trackHeight=150;

        // VERTICAL BOX WITH 2 COMPONENTS
        QBoxLayout *mainVBoxLayout = new QVBoxLayout( this, 2);
        //mainVBoxLayout->setResizeMode(QLayout::FreeResize);

        QSplitter* splitter = new QSplitter(this);
        splitter->setOpaqueResize( false );

                // INSIDE THE SPLITTER : LEFT SIDE
                QString pName1 = "Tracks Panels Area ( Song " + sMyNumber + " )";
                trackPanelsArea = new TrackPanelsArea( this , splitter, (const char*) pName1);// WAS leftBox);

                // INSIDE THE SPLITTER : RIGHT SIDE
                QString pName2 = "Tracks Area ( Song " + sMyNumber + " )";
                tracksArea = new TracksArea( this , splitter, (const char*) pName2 ); // WAS rightBox

        //locator = new Locator( this);
        //locator->show();

        mainVBoxLayout->addWidget(splitter);
        mainVBoxLayout->activate();

        QString projDir = parentProject->get_root_dir();
        QString songNum(""); songNum.setNum(myNumber);
        rootDir = projDir + "/song_" + songNum;
        audioSourcesDir = rootDir + "/audiosources";
        captureDir = rootDir + "/capture";
        mtaProj = rootDir + "/mta.schema"; // mtaProj should be called mtaSchemaFilename


        const char* cRootDir= (const char*) rootDir.latin1();
        if ( mkdir(cRootDir, S_IREAD | S_IWRITE | S_IEXEC)<0)
                {
                if (errno!=EEXIST)
                        {
                        PERROR("Cannot create dir %s" , (const char*) rootDir.latin1());
                        }
                }
        const char* cAudioSourcesDir = (const char*) audioSourcesDir.latin1();
        if (mkdir(cAudioSourcesDir, S_IREAD | S_IWRITE | S_IEXEC)<0)
                {
                if (errno!=EEXIST)
                        {
                        PERROR("Cannot create dir %s ",(const char*) audioSourcesDir.latin1());
                        }
                }
        const char* cCaptureDir = (const char*)captureDir.latin1();
        if (mkdir(cCaptureDir, S_IREAD | S_IWRITE | S_IEXEC)<0)
                {
                if (errno!=EEXIST)
                        {
                        PERROR("Cannot create dir %s",(const char*)  captureDir.latin1());
                        }
                }
        artists = "";
        title = "";
        jogMode = JOG_NONE;
        gainingTrack = -1;
        isScrolling = false;
        jogMode = JOG_NONE;
        hzoom = GlobalProperties::get_int("HZOOM_LEVEL");
        vzoom = 1;
        firstBlock = 0;
        lastBlock = -1;
        workingBlock = 0;
        activeTrackNumber=0;
        takeNumber=0;


        // create the tracks
        for (int i=0; i<numTracks; i++)
                {
                QString sint; sint.setNum(i); sint="Track "+sint;
                track[i]= new Track(this, i * trackHeight, trackHeight);
                trackPanel[i]= new TrackPanel(track[i], i, sint, i * trackHeight, trackHeight);
                track[i]->set_assoc_track_panel(trackPanel[i]);
                }

        shuttleTimer = new QTimer(this);
        connect( shuttleTimer , SIGNAL(timeout()), this, SLOT(update_shuttle()));

        mixer = new Mixer(this);

        cursorManager = new CursorManager(this);
        cursorManager->start();

        QValueList<int> lst;
        lst.append( 45 );
        lst.append( 300 );
        splitter->setSizes( lst );

        filterSelector = new FilterSelector(tracksArea);
        busSelector = new BusSelector(trackPanelsArea);

        trackPanel[activeTrackNumber]->activate();
        track[activeTrackNumber]->activate();

        for (int i=0; i<numTracks; i++)
                {
                track[i]->filterChain->deactivate();
                }

        PEXITCONS;
        }



Song::~Song()
        {
        PENTERDES;
        for (int i=0; i<numTracks; i++)
                {
                delete track[i];
                delete trackPanel[i];
                }
        //delete locator;
        delete cursorManager;
        delete shuttleTimer;
        delete mixer;
        if (audioSourcesList) delete audioSourcesList;
        PEXITDES;
        }




void Song::recreate()
        {
        recreate_tracks();
        recreate_panels();
        show_zoom_info();
        BusMonitor* bm = parentProject->get_bus_monitor();
        bm->clear();
        }


void Song::recreate_tracks()
        {
        cursorManager->disable();
        for (int i=0; i<numTracks; i++)
                track[i]->recreate();
        //if (locator)  locator->clear();
        cursorManager->enable();
        }

void Song::recreate_panels()
        {
        for (int i=0; i<numTracks; i++)
                trackPanel[i]->recreate();
        }

void Song::resizeEvent(QResizeEvent* e)
        {
        //if (locator)  locator->clear();
        recreate();
        }



void Song::mouseMoveEvent(QMouseEvent* e)
        {
        int x = e->x();
        int y = e->y();
        cursorManager->update(x,y);
        if (editingMode==EDITING_MODE_TRACK_CURVES)
                {
                track[activeTrackNumber]->filterChain->followMouse(x,y);
                }

        if (jogMode==JOG_NONE)
                return;
        switch (jogMode)
                {
                case JOG_ZOOM:
                        {
                        update_jog_zoom(e->x());
                        break;
                        }
                case JOG_SHUTTLE:
                        {
                        update_shuttle_factor(e->x());
                        break;
                        }
                case JOG_EDGE:
                        {
                        update_drag_edge(e->x(),e->y());
                        break;
                        }
                case JOG_DRAG:
                        {
                        break;
                        }
                case JOG_GAIN:
                        {
                        update_gain(e->y());
                        trackPanel[gainingTrack]->recreate();
                        break;
                        }
                case JOG_PAN:
                        {
                        update_pan(e->x());
                        trackPanel[gainingTrack]->recreate();
                        break;
                        }
                case JOG_GAIN_AND_PAN:
                        {
                        update_pan(e->x());
                        update_gain(e->y());
                        trackPanel[gainingTrack]->recreate();
                        break;
                        }
                }
        }




int Song::process_jmb_action(MustuxJogMouseBoardMessage* jmbMesg)
        {
        QString actionName = jmbMesg->label;
        int parameter = jmbMesg->parameter;
        int collNumber = jmbMesg->collectedNumber;
        bool isBeginning = jmbMesg->moment;

        if (actionName == "EDITING MODE : NORMAL")
                {
                for (int i=0; i<numTracks; i++)
                        track[i]->set_blur(false);
                editingMode = EDITING_MODE_NORMAL;
                cursorManager->set_mode(CursorManager::MODE_FLOAT);
                cursorManager->set_work_cursor_color(250,50,50);
                for (int i=0; i<numTracks; i++)
                        track[i]->filterChain->deactivate();
                recreate();
                return 1;
                }
        else if (actionName == "EDITING MODE : TRACK CURVES")
                {
                for (int i=0; i<numTracks; i++)
                        track[i]->set_blur(true);
                editingMode = EDITING_MODE_TRACK_CURVES;
                cursorManager->set_mode(CursorManager::MODE_NONE);
                cursorManager->set_work_cursor_color(255,255,255);
                for (int i=0; i<numTracks; i++)
                        track[i]->filterChain->activate();
                recreate();
                return 1;
                }

        // ACTIONS ALLOWED IN ALL EDITING MODES
        if (actionName == "CANCEL")
                {
                // REVISE-ME
                /*
                if (recorder && (recorder->get_status()!=Recorder::REC_NONE))
                        {
                        recorder->abort();
                        info("Recorder aborted!");
                        }
                */
                }
        else if (actionName == "SELECT BUS IN")
                {
                if (isBeginning)
                        {
                        busSelector->start();
                        }
                else
                        {
                        busSelector->stop();
                        int b = busSelector->get_selected_bus();
                        track[activeTrackNumber]->set_bus_in(b);
                        trackPanel[activeTrackNumber]->recreate();
                        }
                }
        if (actionName == "SELECT BUS OUT")
                {
                if (isBeginning)
                        {
                        busSelector->start();
                        }
                else
                        {
                        busSelector->stop();
                        int b = busSelector->get_selected_bus();
                        track[activeTrackNumber]->set_bus_out(b);
                        trackPanel[activeTrackNumber]->recreate();
                        }
                }

        else if (actionName == "TOUCH")
                {
                int trackNumber = collNumber - 1;
                if (trackNumber>=0) // there is a track number collected !!
                        {
                        if ((trackNumber>numTracks) || (trackNumber<0))
                                return -1;
                        touch_track(trackNumber);
                        }
                else
                        touch();
                }
        else if (actionName == "TOUCH AND CENTER")
                {
                int trackNumber = collNumber - 1;
                if (trackNumber>=0) // there is a track number collected !!
                        {
                        if ((trackNumber>numTracks) || (trackNumber<0))
                                return -1;
                        touch_track(trackNumber);
                        }
                else
                        touch();
                center();
                }
        else if (actionName == "HORIZONTAL ZOOM IN")
                {
                hzoom_in();
                }
        else if (actionName == "HORIZONTAL ZOOM OUT")
                {
                hzoom_out();
                }
        else if (actionName == "ZOOM MAGIC KEY")
                {
                jog_zoom();
                }
        else if (actionName == "SCROLL RIGHT")
                {
                scroll_right(100);
                }
        else if (actionName == "SCROLL LEFT")
                {
                scroll_left(100);
                }
        else if (actionName == "SHUTTLE")
                {
                shuttle();
                }
        else if (actionName == "PLAY/STOP")
                {
                toggle_play();
                }
        else if (actionName == "RECORD")
                {
                bool anyTrackArmed = false;
                for (int i=0; i< numTracks; i++)
                        {
                        int r=track[i]->toggle_record();
                        if (r>0)
                                anyTrackArmed=true;
                        }
                if (!anyTrackArmed)
                        info("ARM SOME TRACK BEFORE RECORDING !");
                }
        else if (actionName == "CENTER OF VIEW")
                {
                center();
                }
        else if (actionName == "GO TO BEGIN")
                {
                goto_begin();
                }
        else if (actionName == "MUTE/UNMUTE")
                {
                mute();
                }
        else if (actionName == "SOLO/UNSOLO")
                {
                solo();
                }
        else if (actionName == "TOGGLE TRACK RECORDABLE")
                {
                Track* t = get_track_under_mouse();
                if (t) t->toggle_recordable();
                }
        else if (actionName == "GAIN")
                {
                int trackNumber;
                if (collNumber>=0)
                        {
                        trackNumber = collNumber - 1 ;
                        }
                else
                        {
                        trackNumber = activeTrackNumber;
                        }
                jog_gain(trackNumber);
                }
        else if (actionName == "PAN")
                {
                int trackNumber;
                if (collNumber>=0)
                        {
                        trackNumber = collNumber - 1;
                        }
                else
                        {
                        trackNumber = activeTrackNumber;
                        }
                jog_pan(trackNumber);
                }

        else if (actionName == "GAIN/PAN")
                {
                int track;
                if (collNumber>=0)
                        {
                        track = collNumber*10 + parameter -1;
                        }
                else
                        {
                        track = parameter-1;
                        }
                jog_gain_pan(track);
                }
        else if (actionName == "ADD/INSERT TRACK")
                {
                add_track();
                }


        // NORMAL (CLIP) EDITING MODE RESTRICTED ACTIONS
        if (editingMode == EDITING_MODE_NORMAL)
                {
                if (actionName == "SPLIT")
                        {
                        split();
                        }
                else if (actionName == "SELECT")
                        {
                        select();
                        }
                else if (actionName == "DELETE")
                        {
                        process_delete();
                        }
                else if (actionName == "DRAG AND DROP")
                        {
                        drag(false);
                        }
                else if (actionName == "COPY+DRAG+PASTE")
                        {
                        drag(true);
                        }
                else if (actionName == "JOG-EXPAND/CONTRACT")
                        {
                        drag_edge(0); // [E]
                        }
                else if (actionName == "JOG-EXPAND LEFT")
                        {
                        drag_edge(1); // [EW]
                        }
                else if (actionName == "JOG-EXPAND RIGHT")
                        {
                        drag_edge(2); // [ER]
                        }
                else if (actionName == "JOG-EXPAND LEFT AND RIGHT") 
                        {
                        drag_edge(3); // [WR]
                        }
                else
                        {
                        //info("ACTION NOT ALLOWED IN THIS MODE !");
                        }
                }
        else
        if (editingMode == EDITING_MODE_TRACK_CURVES)
                {
                if (actionName == "ADD FILTER CONTROLLER")
                        {
                        if (isBeginning)
                                {
                                filterSelector->start();
                                }
                        else
                                {
                                filterSelector->stop();
                                QString filterType = filterSelector->get_selected_filter_type();
                                //QString s = "Adding " + filterType + " Controller";
                                //parentProject->get_parent_interface()->defaultLcd->roll_message(s);
                                track[activeTrackNumber]->filterChain->add_filter_controller(filterType);
                                }
                        }
                else if (actionName == "SELECT FILTER CONTROLLER")
                        {
                        track[activeTrackNumber]->filterChain->select_controller(isBeginning);
                        }
                else if (actionName == "REMOVE FILTER CONTROLLER")
                        {
                        track[activeTrackNumber]->filterChain->remove_controller();
                        }

                else if (actionName == "ADD NODE")
                        {
                        track[activeTrackNumber]->filterChain->add_node();
                        }
                else if (actionName == "DRAG AND DROP")
                        {
                        track[activeTrackNumber]->filterChain->drag_node(isBeginning);
                        }
                else if (actionName == "FILTER SETUP")
                        {
                        track[activeTrackNumber]->filterChain->filter_setup();
                        }
                else if (actionName == "NODE SETUP")
                        {
                        track[activeTrackNumber]->filterChain->node_setup();
                        }

                }
        return 1;
        }


MustuxDrawable* Song::get_tracks_area()
        {
        return (MustuxDrawable*) tracksArea;
        }



MustuxDrawable* Song::get_track_panels_area()
        {
        return (MustuxDrawable*) trackPanelsArea;
        }



void Song::set_master_gain(float pMasterGain)
        {
        }


QString Song::get_capture_dir()
        {
        return captureDir;
        }


float Song::get_master_gain()
        {
        return masterGain;
        }


int Song::save()
        {
        PENTER;
        PMESG("Saving Song MTA schema : %s",(const char*) mtaProj.latin1());
        FILE* file = fopen((const char*)mtaProj.ascii(),"w");
        if (!file)
                {
                PERROR("Cannot Save Song #%d" ,myNumber);
                PEXIT;
                return -1;
                }
        QString mta_schema = get_mta_schema();
        char* data = (char*) mta_schema.latin1();
        fwrite( data, 1, mta_schema.length(), file );
        fclose( file );
        PEXIT;
        return 1;
        }



int Song::render()
        {
        PENTER;
        QString filename( QFileDialog::getSaveFileName( title + ".praf", "Praf/Wav Files (*.praf *.wav)", this ) );
        if ( !filename.isEmpty() )
                {
                const char* fn = filename.ascii();
                render(fn);
                PEXIT;
                return 1;
                }
        PEXIT;
        return -1;
        }


int Song::render(QString filename)
        {
        PENTER;
        int r = mixer->render_all(filename);
        if (r<0)
                {
                PERROR("Rendering all tracks failed");
                }
        PEXIT;
        return r;
        }



int Song::load()
        {
        QFile f(mtaProj);
        int ntrack=0;
        if ( f.open(IO_ReadOnly) )
                {
                QTextStream t( &f );
                while (!t.eof())
                        {
                        QString line = t.readLine();
                        line = line.stripWhiteSpace();
                        if (line.find("<track")>=0)
                                {
                                PMESG("Filling track %d ...",ntrack);
                                while ((line.find("</track")<0) && (!t.eof()))
                                        {
                                        line = t.readLine();
                                        line = line.stripWhiteSpace();
                                        if (line.find("<clip")>=0)
                                                {
                                                bool isTake;
                                                int x1 = line.find("source=");
                                                if (x1<0)
                                                        {
                                                        x1 = line.find("take=")+6;
                                                        isTake=true;
                                                        }
                                                else
                                                        {
                                                        x1+=8;
                                                        isTake=false;
                                                        }
                                                int x2 = line.find("\"", x1+1);
                                                QString sSource=line.mid(x1,x2-x1);
                                                int x3 = line.find("TFB=")+5;
                                                int x4 = line.find("\"", x3+1);
                                                QString sTFB=line.mid(x3,x4-x3);
                                                int x5 = line.find("SFB=")+5;
                                                int x6 = line.find("\"", x5+1);
                                                QString sSFB=line.mid(x5,x6-x5);
                                                int x7 = line.find("LEN=")+5;
                                                int x8 = line.find("\"", x7+1);
                                                QString sLEN=line.mid(x7,x8-x7);
                                                long long tfb = (long long) sTFB.toDouble();
                                                long long sfb = (long long) sSFB.toDouble();
                                                long long len = (long long) sLEN.toDouble();
                                                PMESG("Creating clip from [%s] TFB=%d SFB=%d LEN=%d",(const char*) sSource.latin1(),tfb,sfb,len );
                                                Audio* audioSource = audioSourcesList->get_audio_for_source(sSource);
                                                if (!audioSource)
                                                        {
                                                        audioSource = new Audio(this);
                                                        int r = audioSource->build(rootDir+(isTake?"/capture":"/audiosources"),sSource);
                                                        if (r>0)
                                                                {
                                                                audioSourcesList->add(audioSource);
                                                                track[ntrack]->add_clip(audioSource, tfb, sfb, len, isTake);
                                                                }
                                                        }
                                                else
                                                        track[ntrack]->add_clip(audioSource, tfb, sfb, len, isTake);
                                                }
                                        else if (line.find("<filtercontroller")>=0)
                                                {
                                                int x1=line.find("type=")+6;
                                                int x2=line.find("\"",x1);
                                                QString fctype=line.mid(x1,x2-x1);
                                                FilterController* fc = track[ntrack]->filterChain->add_filter_controller(fctype);
                                                while ( fc && (line.find("</filtercontroller")<0) && (!t.eof()))
                                                        {
                                                        line = t.readLine();
                                                        line = line.stripWhiteSpace();
                                                        if (line.find("<curve")>=0)
                                                                {

                                                                int x1=line.find("type=")+6;
                                                                int x2=line.find("\"",x1);
                                                                QString cutype=line.mid(x1,x2-x1);
                                                                PMESG("Adding Curve %s",(const char*)cutype.latin1());
                                                                Curve* cu = fc->add_curve(cutype);
                                                                if (!cu) // curve already exists !
                                                                        cu = fc->get_curve(cutype); // so use the already existent curve

                                                                while (cu && (line.find("</curve")<0) && (!t.eof()))
                                                                        {
                                                                        line = t.readLine();
                                                                        line = line.stripWhiteSpace();
                                                                        if (line.find("<node")>=0)
                                                                                {
                                                                                int x1=line.find("pos=")+4;
                                                                                int x2=line.find(" ",x1);
                                                                                QString spos=line.mid(x1,x2-x1);
                                                                                int x3=line.find("value=")+6;
                                                                                int x4=line.find(" ",x3);
                                                                                QString sval=line.mid(x3,x4-x3);
                                                                                PMESG("Adding node : [%s@%s]",(const char*)sval.latin1(),(const char*)spos.latin1());
                                                                                long long pos= (long long) spos.toDouble();
                                                                                float val= sval.toFloat();
                                                                                CurveNode* no=cu->add_node(pos,val);
                                                                                }
                                                                        }
                                                                }
                                                        }
                                                }

                                        }
                                ntrack++;
                                }
                        }
                }
        if (editingMode!=EDITING_MODE_TRACK_CURVES)
                for (int i=0; i<numTracks; i++)
                        track[i]->filterChain->deactivate();
        return 0;
        }

int Song::import_audio(QString filename)
        {
        PENTER;
        int r=-1;
        if ( !filename.isEmpty() )
                {
                QString originalName = filename;
                QString pureName = originalName.mid(originalName.findRev('/')+1);
                QString targetFilename = audioSourcesDir + "/" + pureName;
                PMESG("Copying %s  to %s", (const char*)originalName.latin1(),(const char*)targetFilename.latin1());
                // cp originalFilename indicated by User to targetFilename inside song's audio sources dir
                FILE* pipe_in;
                QString command = "cp -f \"" + originalName + "\" \"" + targetFilename+"\"";
                pipe_in = popen (command, "r");
                if (!pipe_in)
                        {
                        PERROR("Cannot copy this audio source to audio sources dir");
                        PEXIT;
                        r = -1;
                        }
                else
                        {
                        char c;
                        while (!feof(pipe_in))
                                {
                                fread(&c,1,1,pipe_in);
                                }
                        pclose (pipe_in);
                        if (!create_clip(audioSourcesDir,pureName)) r=-1; else r=0;
                        }
                }
        return r;
        }


Project* Song::get_parent_project()
        {
        return parentProject;
        }



int Song::get_active_track_number()
        {
        return activeTrackNumber;
        }

void Song::touch()
        {
        PENTER;
        touch_x(tracksArea->get_mouse_x());
        PEXIT;
        }

void Song::touch_x(int x)
        {
        PENTER;
        for (int i=0; i<numTracks; i++)
                {
                if ((track[i]->is_pointed()) || (trackPanel[i]->is_pointed()))
                        {
                        workingBlock = xpos_to_block(x);
                        cursorManager->set_work_at(x);
                        PMESG("Track %d touched : Activating it...",i);
                        touch_track(i);
                        break;
                        }
                }
        PEXIT;
        }

void Song::touch_track(int trackNumber)
        {
        track[activeTrackNumber]->deactivate();
        trackPanel[activeTrackNumber]->deactivate();
        activeTrackNumber=trackNumber;
        track[activeTrackNumber]->activate();
        trackPanel[activeTrackNumber]->activate();
        }


void Song::touch_in_active_track_panel() // ? REALLY NECESSARY ??
        {
        PENTER;
        touch_in_active_track_panel(trackPanelsArea->get_mouse_x());
        PEXIT;
        }

void Song::touch_in_active_track_panel(int x) // ? REALLY NECESSARY ??
        {
        PENTER;
// ?
        PEXIT;
        }



void Song::set_active_track(int trackNumber)
        {
        if ((trackNumber>numTracks) || (trackNumber<0))
                return;
        trackPanel[activeTrackNumber]->deactivate();
        track[activeTrackNumber]->deactivate();
        activeTrackNumber=trackNumber;
        trackPanel[activeTrackNumber]->activate();
        track[activeTrackNumber]->activate();
        }


TrackPanel* Song::get_active_track_panel()
        {
        return trackPanel[activeTrackNumber];
        }


TrackPanel* Song::get_track_panel_under_y(int y)
        {
        for (int t=0; t<numTracks; t++)
                {
                if ( (y >= trackPanel[t]->baseY ) && ( y <= trackPanel[t]->baseY + trackPanel[t]->height) )
                        return trackPanel[t];
                }
        return 0;
        }



Track* Song::get_track(int trackNumber)
        {
        if ((trackNumber<0) || (trackNumber>numTracks))
                return (Track*) 0;
        return track[trackNumber];
        }




void Song::jog_gain(int track)
        {
        PENTER;
        if ((track < 0) || (track > numTracks))
                {
                info("INVALID TRACK");
                PEXIT;
                return;
                }
        if (jogMode == JOG_NONE)
                {
                PMESG("Starting jog gain for track %d",track);
                jogMode = JOG_GAIN;
                gainingTrack = track;
                origY = tracksArea->get_mouse_y();
                origGain = trackPanel[track]->get_gain();
                }
        else
                {
                PMESG("Finishing jog gain for track %d",track);
                jogMode = JOG_NONE;
                gainingTrack = -1; // NOT NEEDED .... THIS LINE CAN BE REMOVED
                origGain=0;
                }

        PEXIT;
        }



void Song::jog_pan(int track)
        {
        PENTER;
        if ((track < 0) || (track > numTracks))
                {
                info("INVALID TRACK");
                PEXIT;
                return;
                }
        if (jogMode == JOG_NONE)
                {
                PMESG("Starting jog pan for track %d",track);
                jogMode = JOG_PAN;
                gainingTrack = track;
                origX = tracksArea->get_mouse_x();
                }
        else
                {
                PMESG("Finishing jog pan for track %d",track);
                jogMode = JOG_NONE;
                gainingTrack = -1; // NOT NEEDED .... THIS LINE CAN BE REMOVED
                }

        PEXIT;
        }

void Song::jog_gain_pan(int track)
        {
        PENTER;
        if ((track < 0) || (track > numTracks))
                {
                info("INVALID TRACK");
                PEXIT;
                return;
                }
        if (jogMode == JOG_NONE)
                {
                PMESG("Starting jog gain/pan for track %d",track);
                jogMode = JOG_GAIN_AND_PAN;
                gainingTrack = track;
                origX = tracksArea->get_mouse_x();
                origY = tracksArea->get_mouse_y();
                }
        else
                {
                PMESG("Finishing jog gain/pan for track %d",track);
                jogMode = JOG_NONE;
                gainingTrack = -1; // NOT NEEDED .... THIS LINE CAN BE REMOVED
                }

        PEXIT;
        }




void Song::update_gain(int y)
        {
        float h = (float) height() * 0.6;
        float ofy = (float) origY - y;
        float g = 2.0f * ( ofy / h ) * 12;
        float fg = g + origGain;
        trackPanel[gainingTrack]->set_gain( fg );
        }



void Song::update_pan(int x)
        {
        int w = (int) (width() * 0.6);
        int ofx = origX - x;
/*        if ( abs( ofx ) > 40 )
                {*/
                float p = -2.0f *  (float) (ofx/*-40*/) / w ;
                trackPanel[gainingTrack]->set_pan( p );
/*                }
        if ( abs( ofx ) < -40 )
                {
                float p = 2.0f *  (float) (ofx+40) / w ;
                trackPanel[gainingTrack]->set_pan( p );
                }*/
        }



AudioClip* Song::create_clip(QString audiofileRootdir, QString filename)
        {
        PENTER;
        Audio* audio = new Audio(this);
        if (audio->build(audiofileRootdir,filename)<0)
                {
                PERROR("File not Found : %s%s",(const char*) audiofileRootdir.latin1(),(const char*)  filename.latin1());
                return 0;
                }
        AudioClip* clip = create_clip(audio);
        PEXIT;
        return clip;
        }

AudioClip* Song::create_clip(Audio* au, bool isTake)
        {
        return create_clip(au, isTake, track[activeTrackNumber], workingBlock);
        }
        

AudioClip* Song::create_clip(Audio* au, bool isTake, Track* t, long long insertPos)
        {
        PENTER;
        cursorManager->disable();
        AudioClip* c=0;
        c = t->add_clip(au, insertPos, isTake);
        if (!c)
                {
                PERROR("Cannot add clip into track %d",activeTrackNumber);
                }
        cursorManager->enable();
        recreate_tracks();
        PEXIT;
        return c;
        }


int Song::add_track()
        {
        int nt = numTracks;
        return add_track(nt);
        }


int Song::add_track(int newTrackNumber)
        {
        PENTER;
        // TO ADAPT : track[newTrackNumber]= new Track(this, tracksArea, numTracks * trackHeight,trackHeight);
        // TO ADAPT : trackPanel[newTrackNumber]= new TrackPanel(this, trackPanelsArea, newTrackNumber+1, "Unnamed", numTracks * trackHeight,trackHeight);
        // TO ADAPT : numTracks++;
        PEXIT;
        return 1;
        }




int Song::split()
        {
        PENTER;
        touch();
        cursorManager->disable();
        track[activeTrackNumber]->split_clip(workingBlock);
        cursorManager->enable();
        PEXIT;
        return 1;
        }


int Song::mute()
        {
        PENTER;
        bool validPoint = false;
        int pointedTrack = -1;
        AudioClip* cli = 0 ;
        for (int i=0; i < numTracks; i++)
                {
                if (track[i]->is_pointed())
                        {
                        validPoint = true;
                        pointedTrack = i;
                        }
                }
        if (validPoint)
                {
                int pointedBlock = xpos_to_block(tracksArea->get_mouse_x());
                cli = track[pointedTrack]->get_clip_under(pointedBlock);
                }
        if (cli)
                {
                cursorManager->disable();
                cli->set_muted(!cli->isMuted);
                track[pointedTrack]->recreate();
                cursorManager->enable();
                return 1;
                }
        // no clips pointed ? so check the trackpanels
        for (int i=0; i < numTracks; i++)
                {
                if (trackPanel[i]->is_pointed())
                        {
                        track[i]->isMuted=!track[i]->isMuted; // should be track[i]->toggle_mute();
                        trackPanel[i]->recreate();
                        break;
                        }
                }
        PEXIT;
        return 1;
        }

int Song::solo()
        {
        for (int i=0; i < numTracks; i++)
                {
                if (trackPanel[i]->is_pointed())
                        {
                        track[i]->isSolo=!track[i]->isSolo;
                        trackPanel[i]->recreate();
                        break;
                        }
                }
        return 1;
        }


void Song::set_output_spectrogram(Spectrogram* spectrogram)
        {
        //FOR LATER get_bus_monitor()->set_output_vu(spectrogram);
        }


// PLAY / RECORD STUFF

int Song::toggle_play()
        {
        PENTER;
        if (!mixer->is_playing())
                {
                if (mixer->play(workingBlock)<0)
                        {
                        info("Mixer engine cannot be started");
                        PERROR("Mixer engine cannot be started");
                        PEXIT;
                        return -1;
                        }
                info("PLAYING...");
                cursorManager->disable();
                cursorManager->set_mode(CursorManager::MODE_PLAY);
                cursorManager->enable();
                }
        else
                {
                if (mixer->force_stop()<0)
                        {
                        PERROR("Mixer engine cannot be stopped");
                        PEXIT;
                        return -1;
                        }
                info("STOP");
                }
        PEXIT;
        return 1;
        }




// ZOOMING STUFF

void Song::hzoom_in()
        {
        PENTER;
        if (hzoom > 0)
                {
                hzoom--;
                PMESG("HZOOM Now is %d",hzoom);
                show_zoom_info();
                cursorManager->set_work_at(block_to_xpos(workingBlock));
                recreate_tracks();
                }
        PEXIT;
        }



void Song::set_hzoom_level(int level)
        {
        PENTER;
        if ( ( level > 0) && ( level < Peak::MAX_ZOOM_LEVELS - 1))
                {
                hzoom = level;
                PMESG("HZOOM Now is %d",hzoom);
                show_zoom_info();
                cursorManager->set_work_at(block_to_xpos(workingBlock));
                recreate_tracks();
                }
        PEXIT;
        }


void Song::hzoom_out()
        {
        PENTER;
        if ( hzoom < Peak::MAX_ZOOM_LEVELS - 1  )
                {
                hzoom++;
                PMESG("HZOOM Now is %d",hzoom);
                show_zoom_info();
                cursorManager->set_work_at(block_to_xpos(workingBlock));
                recreate_tracks();
                }
        PEXIT;
        }



void Song::vzoom_out()
        {
        if (vzoom > 1) vzoom--;
        recreate();
        }



void Song::vzoom_in()
        {
        if (vzoom < 5) vzoom++;
        recreate();
        }



void Song::jog_zoom()
        {
        PENTER;
        if (jogMode == JOG_NONE)
                {
                PMESG("Starting jog_zoom : current hzoom =%d",hzoom);
                jogZoomTotalX = width();
                int x = tracksArea->get_mouse_x();
                baseJogZoomFactor = hzoom - ((int) ( (float) x / jogZoomTotalX * 100 / 2 ) + 1 );
                lastJogZoomFactor = -1;
                jogMode = JOG_ZOOM;
                }
        else
                {
                jogMode = JOG_NONE;
                }
        PEXIT;
        }


void Song::update_jog_zoom(int x)
        {
        int jzfactor = (int) ( (float) x / jogZoomTotalX * 100 / 2 ) + 1;
        if (jzfactor==0) jzfactor=1;
        if (jzfactor != lastJogZoomFactor)
                {
                lastJogZoomFactor = jzfactor;
                hzoom = jzfactor + baseJogZoomFactor;
                if ( hzoom < 0 )
                        hzoom = 0;
                if ( hzoom > Peak::MAX_ZOOM_LEVELS -1 )
                        hzoom = Peak::MAX_ZOOM_LEVELS - 1;
                cursorManager->set_work_at(block_to_xpos(workingBlock));
                center();
                recreate_tracks();
                }
        show_zoom_info();
        }


void Song::shuttle()
        {
        PENTER;
        if (jogMode == JOG_NONE)
                {
                PMESG("Starting shuttle");
                jogMode = JOG_SHUTTLE;
                shuttleTimer->start(25);
                }
        else
                {
                PMESG("Stopping shuttle");
                shuttleTimer->stop();
                jogMode = JOG_NONE;
                }
        PEXIT;

        }


void Song::update_shuttle_factor(int x)
        {
        int w = width();
        shuttleFactor = (int) (((float) x / w) * 20 ) - 10;
        }



void Song::update_shuttle()
        {
        if (shuttleFactor < 0)
                scroll_left(shuttleFactor * -20);
        else
                scroll_right(shuttleFactor * 20);
        }


void Song::goto_begin()
        {
        PENTER;
        firstBlock = 0;
        workingBlock = 0;
        cursorManager->set_work_at(block_to_xpos(workingBlock));
        recreate_tracks();
        PEXIT;
        }



void Song::goto_end()
        {
        PENTER;
        //firtBlock = ...
        //for (int i=0; i<MAX_TRACKS; i++)
                //workingBlock = track[i]->get_max_block();
        //center();
        PEXIT;
        }


void Song::center()
        {
        PENTER;
        if (!mixer->is_playing())// && !mixer->isRecording())
                {
                int w = tracksArea->width();
                int half = w/2;
                long long minimumBlockToCenter =  w * Peak::zoomStep[hzoom] / 2;
                if ( workingBlock >= minimumBlockToCenter )
                        {
                        int x = block_to_xpos(workingBlock);
                        if (x<half)
                                scroll_left(half - x);
                        else
                                scroll_right(x-half);
                        }
                recreate_tracks();
                }

        PEXIT;
        }

void Song::scroll_right(int amount)
        {
        PENTER;
        firstBlock += amount*Peak::zoomStep[hzoom];

        // CAN BECOME A recreate_tracks();
        cursorManager->disable();
/*
        for (int i=0; i<numTracks; i++)
                if (track[i]->clipList)
                        track[i]->recreate();
*/
        recreate_tracks();
        //if (locator) locator->clear();
        cursorManager->set_work_at(block_to_xpos(workingBlock));
        cursorManager->enable();
        PEXIT;
        }


void Song::scroll_left(int amount)
        {
        PENTER;
        firstBlock -= amount*Peak::zoomStep[hzoom];
        if (firstBlock<0) firstBlock=0;

        // CAN BECOME A recreate_tracks();
        cursorManager->disable();
/*        for (int i=0; i<numTracks; i++)
                if (track[i]->clipList)
                        track[i]->recreate();
*/
        //if (locator)  locator->clear();
        recreate_tracks();
        cursorManager->set_work_at(block_to_xpos(workingBlock));
        cursorManager->enable();
        PEXIT;
        }







void Song::touch_in_active_track()
        {
        PENTER;
        int x = tracksArea->get_mouse_x();
        touch_in_active_track(x);
        PEXIT;
        }

void Song::touch_in_active_track(int x)
        {
        PENTER;
        workingBlock = xpos_to_block(x);
        cursorManager->set_work_at(x);
        PEXIT;
        }



Track* Song::get_track_under_y(int y)
        {
        for (int t=0; t<numTracks; t++)
                {
                if ( (y >= track[t]->baseY ) && ( y <= track[t]->baseY + track[t]->height) )
                        return track[t];
                }
        return 0;
        }



int Song::select()
        {
        PENTER;
        bool pointingTracks = false;
        int pointedTrack = -1;
        AudioClip* cli = 0 ;
        for (int i=0; i < numTracks; i++)
                {
                if (track[i]->is_pointed())
                        {
                        pointingTracks = true;
                        pointedTrack = i;
                        int x = tracksArea->get_mouse_x();
                        int pointedBlock = xpos_to_block(x);
                        cli = track[pointedTrack]->get_clip_under(pointedBlock);
                        if (cli)
                                {
                                cursorManager->disable();
                                cli->set_selected(!cli->isSelected);
                                track[pointedTrack]->recreate();
                                cursorManager->enable();
                                break;
                                }
                        }
                }
        if (!pointingTracks)
                {
                for (int i=0; i < numTracks; i++)
                        if (trackPanel[i]->is_pointed())
                                {
                                //FOR LATERtrackPanel[i]->toggle_selected();
                                break;
                                }
                }
        return 1;
        PEXIT;
        }


int Song::drag(bool pIsCopying)
        {
        PENTER;
        if ( jogMode == JOG_NONE)
                {
                jogMode = JOG_DRAG;
                isCopying=pIsCopying;
                bool validPoint = false;
                int pointedTrack = -1;
                AudioClip* cli = (AudioClip* ) 0 ;
                for (int i=0; i < numTracks; i++)
                        {
                        if (track[i]->is_pointed())
                                {
                                validPoint = true;
                                pointedTrack = i;
                                }
                        }
                if (validPoint)
                        {
                        int x = tracksArea->get_mouse_x();
                        long long pointedBlock = xpos_to_block(x); //oops shouldnt it be long long ?
                        cli = track[pointedTrack]->get_clip_under(pointedBlock);
                        }
                if (cli)
                        {
                        workingClip = cli;
                        int cx1 = block_to_xpos( cli->trackFirstBlock );
                        int cx2 = block_to_xpos( cli->trackLastBlock );
                        int cy1 = track[pointedTrack]->baseY;
                        int origX = tracksArea->get_mouse_x();
                        int origY = tracksArea->get_mouse_y();
                        cursorManager->disable();
                        cursorManager->set_mode(CursorManager::MODE_DRAG,
                                                origX,
                                                origY,
                                                cx2 - cx1,
                                                track[pointedTrack]->height - 5,
                                                origX - cx1,
                                                origY - cy1
                                                );
                        cursorManager->enable();
                        }
                else
                        {
                        workingClip = (AudioClip*) 0;
                        }

                }
        else
                {
                if (workingClip)
                        {
                        cursorManager->disable();
                        int x = tracksArea->get_mouse_x();
                        int y = tracksArea->get_mouse_y();
                        long long newInsertBlock = xpos_to_block( x - cursorManager->get_x_offset());
                        Track* originTrack = workingClip->parentTrack;
                        Track* targetTrack = get_track_under_y(y);
                        if (!targetTrack)
                                {
                                if (!isCopying)
                                        {
                                        PMESG("Deleting clip %p",workingClip);
                                        originTrack->delete_clip(workingClip);
                                        originTrack->recreate();
                                        }
                                }
                        else
                                {
                                targetTrack->add_clip(workingClip->audioSource, newInsertBlock, workingClip->sourceFirstBlock, workingClip->get_lenght(), workingClip->isTake);
                                targetTrack->recreate();
                                if (!isCopying)
                                        {
                                        originTrack->delete_clip(workingClip);
                                        originTrack->recreate();
                                        }
                                }
                        cursorManager->set_mode(CursorManager::MODE_FLOAT);
                        cursorManager->enable();
                        }
                jogMode = JOG_NONE;
                workingClip=(AudioClip*) 0;
                }
        PEXIT;
        return true;
        }



int Song::drag_edge(int mode)
        {
        PENTER;
        if (jogMode ==  JOG_NONE)
                {
                bool validPoint = false;
                int pointedTrack = -1;
                AudioClip* cli = 0 ;
                for (int i=0; i < numTracks; i++)
                        {
                        if (track[i]->is_pointed())
                                {
                                validPoint = true;
                                pointedTrack = i;
                                }
                        }
                if (validPoint)
                        {
                        int pointedBlock = xpos_to_block(tracksArea->get_mouse_x());
                        cli = track[pointedTrack]->get_clip_under(pointedBlock);
                        }
                if (cli)
                        {
                        jogMode = JOG_EDGE;
                        workingClip = cli;
                        int x = tracksArea->get_mouse_x();
                        if (mode==0) // [E]
                                {
                                int cxm = block_to_xpos( cli->trackFirstBlock + (cli->get_lenght()/2));
                                if (x < cxm)
                                        draggingEdge = LEFT_EDGE;
                                else
                                        draggingEdge = RIGHT_EDGE;
                                }
                        else if (mode==1) draggingEdge = LEFT_EDGE; // [EW]
                        else if (mode==2) draggingEdge = RIGHT_EDGE; // [ER]
                        else if (mode==3) draggingEdge = BOTH_EDGES; // [WR]
                        if (draggingEdge==RIGHT_EDGE)
                                {
                                cursorManager->set_x_offset(block_to_xpos( cli->trackLastBlock ) - x);
                                }
                        else
                                {
                                cursorManager->set_x_offset(block_to_xpos( cli->trackFirstBlock ) - x);
                                }
                        }
                }
        else
                {
                cursorManager->disable();
                track[activeTrackNumber]->recreate();
                cursorManager->enable();
                jogMode = JOG_NONE;
                }
        PEXIT;
        return true;
        }



void Song::update_drag_edge(int x, int y)
        {
        long long b = xpos_to_block(x + cursorManager->get_x_offset());
        long long tfb = workingClip->trackFirstBlock;
        long long tlb = workingClip->trackLastBlock;
        long long sfb = workingClip->sourceFirstBlock;
        long long slb = workingClip->sourceLastBlock;
        long long totb = workingClip->audioSource->get_total_blocks();
        bool recr = false;
        if (draggingEdge==LEFT_EDGE)
                {
                if (sfb + (b-tfb)>=0)
                        {
                        workingClip->set_left_edge(b);
                        recr = true;
                        }
                }
        else
        if (draggingEdge==RIGHT_EDGE)
                {
                if (slb + (b-tlb) <= totb )
                        {
                        workingClip->set_right_edge(b);
                        recr = true;
                        }
                }
        else
        if (draggingEdge==BOTH_EDGES)
                {
                long long w = tlb-tfb;
                if ((sfb + (b-tfb)>=0) && (slb + (b+w-tlb) <= totb ))
                        {
                        workingClip->set_left_edge(b);
                        workingClip->set_right_edge(b+w);
                        recr = true;
                        }
                }
        if (recr)
                {
                cursorManager->disable();
                workingClip->parentTrack->recreate();// same as track[pointedtrack]->recreate(); maybe could use this second since is more clear
                cursorManager->enable();
                }
        }


/*int Song::add_track_env_node(QString whichEnvCurve, int x, int y)
        {
        Track* t = get_track_under_y (y);
        long long pos = xpos_to_block(x);
        float val = y - t->baseY ;
        if (whichEnvCurve == "GAIN")
                t->add_gain_node( pos, val );
        else
        if (whichEnvCurve == "PAN")
                t->add_pan_node( pos, val );
        }

*/

int Song::process_delete()
        {
        PENTER;
        track[activeTrackNumber]->deactivate();
        AudioClip* cli = 0 ;
        for (int i=0; i < numTracks; i++)
                {
                bool someClipDeleted=false;
                cli=track[i]->clipList;
                while (cli)
                        {
                        if (cli->isSelected)
                                {
                                AudioClip* c = cli->next;
                                track[i]->delete_clip(cli);
                                cli=c;
                                someClipDeleted=true;
                                }
                        else
                                cli=cli->next;
                        }
                if (someClipDeleted)
                        {
                        recreate_tracks();
                        }
                }
        track[activeTrackNumber]->activate();
        return 1;
        PEXIT;
        }



int Song::block_to_xpos(long long block)
        {
        int x= ((block - firstBlock)  / Peak::zoomStep[hzoom]);
        return x;
        }



long long Song::xpos_to_block(int xpos)
        {
        long long b = firstBlock + xpos * Peak::zoomStep[hzoom];
        return b;
        }




// get / set stuff
int Song::get_hzoom()
        {
        return hzoom;
        }


int Song::get_vzoom()
        {
        return vzoom;
        }

int Song::get_zoom_step()
        {
        return Peak::zoomStep[hzoom];
        }


long long Song::get_first_block()
        {
        return firstBlock;
        }


long long Song::get_last_block()
        {
        return lastBlock;
        }


void Song::set_first_block(long long pos)
        {
        firstBlock = (pos >=0) ? pos : 0;
        lastBlock = xpos_to_block(width());
        }

int Song::get_playing_xpos()
        {
        long long lo = mixer->get_current_playing_block();
        int pxp = block_to_xpos(lo - PLAY_BLOCKS_OFFSET);
        return pxp;
        }

long long int Song::get_playing_block()
        {
        return mixer->get_current_playing_block();
        }

void Song::disable_cursors()
        {
        cursorManager->disable();
        }


void Song::enable_cursors()
        {
        cursorManager->enable();
        }


QString Song::get_mta_schema()
        {
        QString output="";
        for (int i=0; i < numTracks; i++)
                {
                output = output + track[i]->get_schema();
                }
        return output;
        }



int Song::get_total_tracks()
        {
        return numTracks;
        }


int Song::get_mouse_x()
        {
        return mapFromGlobal(QCursor::pos()).x();
        }


int Song::get_mouse_y()
        {
        return mapFromGlobal(QCursor::pos()).y();
        }

MustuxLcd* Song::get_default_lcd()
        {
        return parentProject->get_default_lcd();
        }

// This method is here just to avoid the annoying warning "QObject::connect: No such slot Song::cleanupEventFilter()"
// when running protux. If anybody knows how to avoid this without implementing this method, please, send an email to
// one of protux developers. Thanks a lot.
void Song::cleanupEventFilter()
        {
        }


void Song::info(QString message)
        {
        parentProject->get_parent_interface()->info(message);
        }

void Song::info(QString message, int line)
        {
        parentProject->get_parent_interface()->info(message, line);
        }

void Song::show_zoom_info()
        {
        MustuxLcd* lcd = get_default_lcd();
        lcd->clear(500,0,160,70);
        QPainter* p = lcd->painter;
        p->setPen(QColor(0,0,20));
        p->setFont(QFont("Fixed",10));
        QString sHzoom; sHzoom.setNum(Peak::zoomStep[hzoom]); sHzoom = "1:" + sHzoom;
        if (hzoom <= Peak::MAX_ZOOM_USING_SOURCEFILE)
                {
                sHzoom = sHzoom + " Micro View";
                }
        p->drawText(510,25,"HZ " + sHzoom);
        p->drawText(510,45,"VZ x1");
        lcd->update(500,0,160,70);
        }

int Song::get_editing_mode()
        {
        return editingMode;
        }

        
int Song::get_status()
        {
        if (mixer->is_playing())
                return PLAYING;
        
        for (int i=0; i<numTracks; i++)
                {
                if (track[i]->get_recorder_status()==Recorder::REC_ARMED)
                        {
                        return ARMED;
                        }
                if (track[i]->get_recorder_status()==Recorder::REC_RECORDING)
                        {
                        return RECORDING;
                        }
                }
        // TODO : check if it is rendering to return RENDERING
        return STOPPED;
        }

long long Song::get_working_block()
        {
        return workingBlock;
        }
        
        
long long Song::get_last_song_block()
        {
        // find the last blockpos in MTA
        int numTracks = get_total_tracks();
        long long lastBlockPos=0;
        for (int i=0; i<numTracks; i++)
                {
                AudioClip* clip = track[i]->clipList;
                while (clip != 0)
                        {
                        long long borderR = clip->trackFirstBlock + (clip->sourceLastBlock - clip->sourceFirstBlock);
                        if (borderR>=lastBlockPos)
                                lastBlockPos = borderR;
                        clip=clip->next;
                        }
                }
        return lastBlockPos;
        }

BusMonitor* Song::get_bus_monitor()
        {
        return parentProject->get_parent_interface()->get_bus_monitor();
        }


Track* Song::get_track_under_mouse()
        {
        for (int i=0; i < numTracks; i++)
                {
                if ((trackPanel[i]->is_pointed()) || (track[i]->is_pointed()))
                        return track[i];
                }
        return (Track*) 0;
        }

//eof

